from os.path import join, dirname
from cffi import FFI
from cffi.api import CDefError
from itertools import chain

sq_inc = join(dirname(__file__), '../../../include/sequoia')
pgp_inc = join(dirname(__file__), '../../../../openpgp-ffi/include/sequoia')
defs = "".join(l
               for l in chain(open(join(pgp_inc, "openpgp/error.h")).readlines(),
                              open(join(pgp_inc, "io.h")).readlines(),
                              open(join(pgp_inc, "openpgp/types.h")).readlines(),
                              open(join(pgp_inc, "openpgp/crypto.h")).readlines(),
                              open(join(pgp_inc, "openpgp/packet.h")).readlines(),
                              open(join(pgp_inc, "openpgp/serialize.h")).readlines(),
                              open(join(pgp_inc, "openpgp.h")).readlines(),
                              open(join(sq_inc, "core.h")).readlines(),
                              open(join(sq_inc, "net.h")).readlines(),
                              open(join(sq_inc, "store.h")).readlines())
               if not l.startswith('#'))

defs = defs.replace("INT_MAX", "{}".format(1<<31))

ffibuilder = FFI()
ffibuilder.set_source('_sequoia',
                      '#include <sequoia.h>',
                      libraries=['sequoia_ffi'])

# cffi magic to make time_t work.
ffibuilder.cdef('typedef int... time_t;')

# free(3)
ffibuilder.cdef('void free (void *ptr);')

try:
    ffibuilder.cdef(defs, override=True)
except CDefError as e:
    try:
        current_decl = e.args[1]
        linenum = current_decl.coord.line

        print("ATTENTION:  Line numbers are not very reliable :(")
        print()
        print("{}: {}".format(linenum, e.args[0]))
        print()

        for i, l in enumerate(defs.split("\n")):
            if i < linenum - 5 or i > linenum + 5:
                continue
            print("{}: {}".format(i + 1, l))
        print()

        print("AST:", e.args[1])
        print()
    except (AttributeError, TypeError, IndexError):
        pass

    raise

if __name__ == '__main__':
    ffibuilder.compile(verbose=True)



from enum import Enum

from _sequoia import ffi, lib
from .error import Error
from .glue import _str, SQObject, invoke
from .core import AbstractReader, AbstractWriter

class KeyID(SQObject):
    _del = lib.pgp_keyid_free
    _clone = lib.pgp_keyid_clone
    _str = lib.pgp_keyid_to_string
    _debug = lib.pgp_keyid_debug
    _eq = lib.pgp_keyid_equal
    _hash = lib.pgp_keyid_hash

    @classmethod
    def from_bytes(cls, fp):
        if len(fp) != 8:
            raise Error("KeyID must be of length 8")
        return KeyID(lib.pgp_keyid_from_bytes(
            ffi.cast("uint8_t *", ffi.from_buffer(fp))))

    @classmethod
    def from_hex(cls, fp):
        return KeyID(lib.pgp_keyid_from_hex(fp.encode()))

    def hex(self):
        return _str(lib.pgp_keyid_to_hex(self.ref()))

class Fingerprint(SQObject):
    _del = lib.pgp_fingerprint_free
    _clone = lib.pgp_fingerprint_clone
    _str = lib.pgp_fingerprint_to_string
    _debug = lib.pgp_fingerprint_debug
    _eq = lib.pgp_fingerprint_equal
    _hash = lib.pgp_fingerprint_hash

    @classmethod
    def from_bytes(cls, fp):
        return Fingerprint(lib.pgp_fingerprint_from_bytes(
            ffi.cast("uint8_t *", ffi.from_buffer(fp)), len(fp)))

    @classmethod
    def from_hex(cls, fp):
        return Fingerprint(lib.pgp_fingerprint_from_hex(fp.encode()))

    def hex(self):
        return _str(lib.pgp_fingerprint_to_hex(self.ref()))

    def keyid(self):
        return KeyID(lib.pgp_fingerprint_to_keyid(self.ref()))

class PacketPile(SQObject):
    _debug = lib.pgp_packet_pile_debug
    _del = lib.pgp_packet_pile_free
    _clone = lib.pgp_packet_pile_clone

    @classmethod
    def from_reader(cls, ctx, reader):
        return PacketPile(invoke(lib.pgp_packet_pile_from_reader, reader.ref()),
                          context=ctx)

    @classmethod
    def open(cls, ctx, filename):
        return PacketPile(invoke(lib.pgp_packet_pile_from_file, filename.encode()),
                          context=ctx)

    @classmethod
    def from_bytes(cls, ctx, source):
        return PacketPile(invoke(lib.pgp_packet_pile_from_bytes,
                                 ffi.from_buffer(source),
                                 len(source)),
                          context=ctx)

    def serialize(self, writer):
        status = invoke(lib.pgp_packet_pile_serialize,
                        self.ref(),
                        writer.ref())
        if status:
            raise Error._last(self.context())

class Cert(SQObject):
    _del = lib.pgp_cert_free
    _clone = lib.pgp_cert_clone
    _eq = lib.pgp_cert_equal
    _str = lib.pgp_cert_to_string
    _debug = lib.pgp_cert_debug

    @classmethod
    def from_reader(cls, ctx, reader):
        return Cert(invoke(lib.pgp_cert_from_reader, reader.ref()),
                   context=ctx)

    @classmethod
    def open(cls, ctx, filename):
        return Cert(invoke(lib.pgp_cert_from_file, filename.encode()),
                   context=ctx)

    @classmethod
    def from_packet_pile(cls, ctx, packet_pile):
        return Cert(invoke(lib.pgp_cert_from_packet_pile, packet_pile.ref_consume()),
                   context=ctx)

    @classmethod
    def from_bytes(cls, ctx, source):
        return Cert(invoke(lib.pgp_cert_from_bytes,
                          ffi.from_buffer(source),
                          len(source)),
                   context=ctx)

    def serialize(self, writer):
        status = invoke(lib.pgp_cert_serialize,
                        self.ref(),
                        writer.ref())
        if status:
            raise Error._last(self.context())

    def fingerprint(self):
        return Fingerprint(lib.pgp_cert_fingerprint(self.ref()),
                           context=self.context())

    def merge(self, other):
        new = invoke(lib.pgp_cert_merge,
                     self.ref_consume(),
                     other.ref_consume())
        if new == ffi.NULL:
            raise Error._last(self.context())
        self.ref_replace(new)

class Kind(Enum):
    Message = lib.PGP_ARMOR_KIND_MESSAGE
    PublicKey = lib.PGP_ARMOR_KIND_PUBLICKEY
    SecretKey = lib.PGP_ARMOR_KIND_SECRETKEY
    Signature = lib.PGP_ARMOR_KIND_SIGNATURE
    File = lib.PGP_ARMOR_KIND_FILE
    Any = lib.PGP_ARMOR_KIND_ANY

class ArmorReader(AbstractReader):
    @classmethod
    def new(cls, ctx, inner, kind=Kind.Any):
        ar = ArmorReader(lib.pgp_armor_reader_new(inner.ref(),
                                                 kind.value),
                         context=ctx)
        ar.inner = inner
        return ar

    def close(self):
        super(ArmorReader, self)._delete()
        self.inner.close()

class ArmorWriter(AbstractWriter):
    @classmethod
    def new(cls, ctx, inner, kind):
        aw = ArmorWriter(invoke(lib.pgp_armor_writer_new,
                                inner.ref(),
                                kind.value,
                                ffi.NULL, 0), #XXX headers
                         context=ctx)
        aw.inner = inner
        return aw

    def close(self):
        invoke(lib.pgp_armor_writer_finalize, self.ref())
        super(ArmorWriter, self)._delete(skip_free=True)
        self.inner.close()

class Tag(Enum):
    PKESK = lib.PGP_TAG_PKESK
    Signature = lib.PGP_TAG_SIGNATURE
    SKESK = lib.PGP_TAG_SKESK
    OnePassSig = lib.PGP_TAG_ONE_PASS_SIG
    SecretKey = lib.PGP_TAG_SECRET_KEY
    PublicKey = lib.PGP_TAG_PUBLIC_KEY
    SecretSubkey = lib.PGP_TAG_SECRET_SUBKEY
    CompressedData = lib.PGP_TAG_COMPRESSED_DATA
    SED = lib.PGP_TAG_SED
    Marker = lib.PGP_TAG_MARKER
    Literal = lib.PGP_TAG_LITERAL
    Trust = lib.PGP_TAG_TRUST
    UserID = lib.PGP_TAG_USER_ID
    PublicSubkey = lib.PGP_TAG_PUBLIC_SUBKEY
    Unassigned15 = lib.PGP_TAG_UNASSIGNED15
    Unassigned16 = lib.PGP_TAG_UNASSIGNED16
    UserAttribute = lib.PGP_TAG_USER_ATTRIBUTE
    SEIP = lib.PGP_TAG_SEIP
    MDC = lib.PGP_TAG_MDC
    # xxx the rest

class Key(SQObject):
    @property
    def fingerprint(self):
        return Fingerprint(lib.pgp_key_fingerprint(self.ref()))

    @property
    def keyid(self):
        return KeyID(lib.pgp_key_keyid(self.ref()))

class PublicKey(Key):
    pass
class PublicSubkey(Key):
    pass
class SecretKey(Key):
    pass
class SecretSubkey(Key):
    pass

class UserID(SQObject):
    @property
    def value(self):
        buf_len = ffi.new("size_t[1]")
        buf = lib.pgp_user_id_value(self.ref(), buf_len)
        return ffi.buffer(buf, buf_len[0])

class UserAttribute(SQObject):
    @property
    def value(self):
        buf_len = ffi.new("size_t[1]")
        buf = lib.pgp_user_attribute_value(self.ref(), buf_len)
        return ffi.buffer(buf, buf_len[0])

class SKESK(SQObject):
    def decrypt(self, passphrase):
        algo = ffi.new("uint8_t[1]")
        key = ffi.new("uint8_t[32]")
        key_len = ffi.new("size_t[1]")
        key_len[0] = len(key)
        invoke(lib.pgp_skesk_decrypt,
               self.ref(),
               ffi.from_buffer(passphrase),
               len(passphrase),
               algo, key, key_len)
        return (algo[0], ffi.buffer(key, key_len[0]))

class SEIP(SQObject):
    pass

class Packet(SQObject):
    _map = {
        Tag.PublicKey: lambda x, **kwargs: PublicKey(x, **kwargs),
        Tag.PublicSubkey: lambda x, **kwargs: PublicSubkey(x, **kwargs),
        Tag.SecretKey: lambda x, **kwargs: SecretKey(x, **kwargs),
        Tag.SecretSubkey: lambda x, **kwargs: SecretSubkey(x, **kwargs),
        Tag.UserID: lambda x, **kwargs: UserID(x, **kwargs),
        Tag.UserAttribute: lambda x, **kwargs: UserAttribute(x, **kwargs),
        Tag.SKESK: lambda x, **kwargs: SKESK(x, **kwargs),
        Tag.SEIP: lambda x, **kwargs: SEIP(x, **kwargs),
    }
    @property
    def tag(self):
        return Tag(lib.pgp_packet_tag(self.ref()))
    @property
    def kind(self):
        return Tag(lib.pgp_packet_kind(self.ref()))
    def __str__(self):
        return "<Packet tag={}>".format(self.tag)
    def match(self):
        return self._map[self.kind](self.ref(), context=self.context(), owner=self)

class PacketParserResult(SQObject):
    _del = lib.pgp_packet_parser_result_free

    def packet_parser(self):
        ref = lib.pgp_packet_parser_result_packet_parser(self.ref())
        if ref != ffi.NULL:
            # Success!  We are consumed.
            self._delete(skip_free=True)
            return PacketParser(ref, context=self.context())
        else:
            return None

    def eof(self):
        ref = lib.pgp_packet_parser_result_eof(self.ref())
        if ref != ffi.NULL:
            # Success!  We are consumed.
            self._delete(skip_free=True)
            return PacketParserEOF(ref, context=self.context())
        else:
            return None

class PacketParserEOF(SQObject):
    _del = lib.pgp_packet_parser_eof_free

    def is_message(self):
        return invoke(lib.pgp_packet_parser_eof_is_message, self.ref())

class PacketParser(SQObject):
    _del = lib.pgp_packet_parser_free

    @classmethod
    def from_reader(cls, ctx, reader):
        return PacketParserResult(
            invoke(lib.pgp_packet_parser_from_reader, reader.ref()),
            context=ctx)

    @classmethod
    def open(cls, ctx, filename):
        return PacketParserResult(
            invoke(lib.pgp_packet_parser_from_file, filename.encode()),
            context=ctx)

    @classmethod
    def from_bytes(cls, ctx, source):
        return PacketParserResult(
            invoke(lib.pgp_packet_parser_from_bytes,
                   ffi.from_buffer(source),
                   len(source)),
            context=ctx)

    @property
    def has_packet(self):
        return self.ref() != ffi.NULL

    @property
    def packet(self):
        return Packet(lib.pgp_packet_parser_packet(self.ref()),
                      context=self.context(),
                      owner=self)

    @property
    def recursion_depth(self):
        return lib.pgp_packet_parser_recursion_depth(self.ref())

    def next(self):
        packet = ffi.new("pgp_packet_t[1]")
        old_rl = ffi.new("uint8_t[1]")
        ppr = ffi.new("pgp_packet_parser_result_t[1]")
        new_rl = ffi.new("uint8_t[1]")

        invoke(lib.pgp_packet_parser_next,
               self.ref_consume(),
               packet,
               ppr)

        return (Packet(packet[0]), PacketParserResult(ppr[0], self.context()))

    def recurse(self):
        packet = ffi.new("pgp_packet_t[1]")
        old_rl = ffi.new("uint8_t[1]")
        ppr = ffi.new("pgp_packet_parser_result_t[1]")
        new_rl = ffi.new("uint8_t[1]")

        invoke(lib.pgp_packet_parser_recurse,
               self.ref_consume(),
               packet,
               ppr)

        return (Packet(packet[0]), PacketParserResult(ppr[0], self.context()))

    def buffer_unread_content(self):
        buf_len = ffi.new("size_t[1]")
        buf = invoke(lib.pgp_packet_parser_buffer_unread_content,
                     self.ref(),
                     buf_len)
        return ffi.buffer(buf, buf_len[0])

    def decrypt(self, algo, key):
        invoke(lib.pgp_packet_parser_decrypt,
               self.ref(),
               algo,
               ffi.from_buffer(key),
               len(key))



from _sequoia import ffi, lib

from .error import Error
from .glue import _str, _static_str, SQObject, sq_iterator, sq_time
from .net import NetworkPolicy
from .openpgp import Fingerprint, Cert

class Store(object):
    @classmethod
    def log(cls, ctx):
        yield from sq_iterator(
            ffi.gc(
                lib.sq_store_server_log(ctx.ref()),
                lib.sq_log_iter_free),
            lib.sq_log_iter_next,
            lambda x: Log(x, context=ctx))

    @classmethod
    def list_keys(cls, ctx):
        def next_fn(i):
            fpp = ffi.new("pgp_fingerprint_t[1]")
            key = lib.sq_key_iter_next(i, fpp)
            if key == ffi.NULL:
                return ffi.NULL
            else:
                return (Fingerprint(fpp[0], ctx),
                        Key(key, ctx))

        yield from sq_iterator(
            ffi.gc(
                lib.sq_store_list_keys(ctx.ref()),
                lib.sq_key_iter_free),
            next_fn)

class Mapping(SQObject):
    _del = lib.sq_mapping_free

    # Keys used for communications.
    REALM_CONTACTS = _static_str(lib.SQ_REALM_CONTACTS)

    # Keys used for signing software updates.
    REALM_SOFTWARE_UPDATES = _static_str(lib.SQ_REALM_SOFTWARE_UPDATES)

    @classmethod
    def open(cls, ctx, network_policy=NetworkPolicy.Encrypted, realm=REALM_CONTACTS, name="default"):
        return Mapping(lib.sq_mapping_open(ctx.ref(), network_policy.value,
                                        realm.encode(), name.encode()),
                        context=ctx)


    def add(self, label, fingerprint):
        return Binding(lib.sq_mapping_add(self.context().ref(), self.ref(),
                                        label.encode(), fingerprint.ref()),
                       context=self.context())

    def import_(self, label, cert):
        return Cert(lib.sq_mapping_import(self.context().ref(), self.ref(),
                                       label.encode(), cert.ref()),
                   context=self.context())

    def lookup(self, label):
        return Binding(lib.sq_mapping_lookup(self.context().ref(), self.ref(),
                                           label.encode()),
                       self.context())

    def delete(self):
        if lib.sq_mapping_delete(self.ref()):
            raise Error._last(self.context())
        super(Mapping, self)._delete(skip_free=True)

    def iter(self):
        def next_fn(i):
            labelp = ffi.new("char *[1]")
            fpp = ffi.new("pgp_fingerprint_t[1]")
            binding = lib.sq_binding_iter_next(i, labelp, fpp)
            if binding == ffi.NULL:
                return ffi.NULL
            else:
                return (_str(labelp[0]),
                        Fingerprint(fpp[0], self.context()),
                        Binding(binding, self.context()))

        yield from sq_iterator(
            ffi.gc(
                lib.sq_mapping_iter(self.context().ref(), self.ref()),
                lib.sq_binding_iter_free),
            next_fn)

    def log(self):
        yield from sq_iterator(
            ffi.gc(
                lib.sq_mapping_log(self.context().ref(), self.ref()),
                lib.sq_log_iter_free),
            lib.sq_log_iter_next,
            lambda x: Log(x, context=self.context()))

class Binding(SQObject):
    _del = lib.sq_binding_free

    def stats(self):
        return Stats(lib.sq_binding_stats(self.context().ref(), self.ref()),
                     self.context())

    def key(self):
        return Key(lib.sq_binding_key(self.context().ref(), self.ref()),
                   self.context())

    def cert(self):
        return Cert(lib.sq_binding_cert(self.context().ref(), self.ref()),
                   self.context())

    def import_(self, cert):
        return Cert(lib.sq_binding_import(self.context().ref(), self.ref(), cert),
                   self.context())

    def rotate(self, cert):
        return Cert(lib.sq_binding_rotate(self.context().ref(), self.ref(), cert),
                   self.context())

    def delete(self):
        if lib.sq_binding_delete(self.ref()):
            raise Error._last(self.context())
        super(Binding, self)._delete(skip_free=True)

    def log(self):
        yield from sq_iterator(
            ffi.gc(
                lib.sq_binding_log(self.context().ref(), self.ref()),
                lib.sq_log_iter_free),
            lib.sq_log_iter_next,
            lambda x: Log(x, context=self.context()))

class Key(SQObject):
    _del = lib.sq_key_free

    def stats(self):
        return Stats(lib.sq_key_stats(self.context().ref(), self.ref()),
                     self.context())

    def cert(self):
        return Cert(lib.sq_key_cert(self.context().ref(), self.ref()),
                   self.context())

    def import_(self, cert):
        return Cert(lib.sq_key_import(self.context().ref(), self.ref(), cert),
                   self.context())

    def log(self):
        yield from sq_iterator(
            ffi.gc(
                lib.sq_key_log(self.context().ref(), self.ref()),
                lib.sq_log_iter_free),
            lib.sq_log_iter_next)


class Stats(SQObject):
    _del = lib.sq_stats_free
    def __init__(self, o, context=None):
        super(Stats, self).__init__(o, context=context)
        self.encryption = Stamps(ffi.addressof(o, "encryption"))
        self.verification = Stamps(ffi.addressof(o, "verification"))

    @property
    def created(self):
        return sq_time(self.ref().created)

    @property
    def updated(self):
        return sq_time(self.ref().updated)

    def __str__(self):
        return \
            "Stats{{created={}, updated={}, encryption={}, verification={}}}" \
            .format(self.created, self.updated, self.encryption,
                    self.verification)

class Stamps(SQObject):
    @property
    def count(self):
        return self.ref().count

    @property
    def first(self):
        return sq_time(self.ref().first)

    @property
    def last(self):
        return sq_time(self.ref().last)

    def __str__(self):
        return "Stamps{{count={}, first={}, last={}}}".format(
            self.count, self.first, self.last)

class Log(SQObject):
    _del = lib.sq_log_free

    @property
    def timestamp(self):
        return sq_time(self.ref().timestamp)

    @property
    def store(self):
        if self.ref().store == ffi.NULL:
            return None
        else:
            return Store(self.ref().store, context=self.context(),
                         owner=self)

    @property
    def binding(self):
        if self.ref().binding == ffi.NULL:
            return None
        else:
            return Binding(self.ref().binding, context=self.context(),
                           owner=self)

    @property
    def key(self):
        if self.ref().key == ffi.NULL:
            return None
        else:
            return Key(self.ref().key, context=self.context(),
                       owner=self)

    @property
    def slug(self):
        return ffi.string(self.ref().slug).decode()

    @property
    def status(self):
        return ffi.string(self.ref().status).decode()

    @property
    def error(self):
        if self.ref().error == ffi.NULL:
            return None
        else:
            return ffi.string(self.ref().error).decode()

    def __str__(self):
        if self.error:
            return "{}: {}: {}: {}".format(
                self.timestamp, self.slug, self.status, self.error)
        else:
            return "{}: {}: {}".format(
                self.timestamp, self.slug, self.status)



from enum import Enum

from _sequoia import ffi, lib
from .openpgp import Cert
from .error import Error
from .glue import SQObject

class NetworkPolicy(Enum):
    Offline = lib.SQ_NETWORK_POLICY_OFFLINE
    Anonymized = lib.SQ_NETWORK_POLICY_ANONYMIZED
    Encrypted = lib.SQ_NETWORK_POLICY_ENCRYPTED
    Insecure = lib.SQ_NETWORK_POLICY_INSECURE

class KeyServer(SQObject):
    _del = lib.sq_keyserver_free

    @classmethod
    def new(cls, ctx, uri, network_policy=NetworkPolicy.Encrypted, cert=None):
        if not cert:
            ks = lib.sq_keyserver_new(
                    ctx.ref(),
                    network_policy.value,
                    uri.encode())
        else:
            ks = lib.sq_keyserver_with_cert(
                ctx.ref(),
                network_policy.value,
                uri.encode(),
                ffi.cast("uint8_t *", ffi.from_buffer(cert)),
                len(cert))
        return KeyServer(ks, context=ctx)

    @classmethod
    def keys_openpgp_org(cls, ctx, network_policy=NetworkPolicy.Encrypted):
        ks = lib.sq_keyserver_keys_openpgp_org(ctx.ref(),
                                    network_policy.value)
        return KeyServer(ks, context=ctx)

    def get(self, keyid):
        return Cert(lib.sq_keyserver_get(self.context().ref(),
                                        self.ref(),
                                        keyid.ref()),
                   context=self.context())

    def send(self, cert):
        r = lib.sq_keyserver_send(self.context().ref(),
                                  self.ref(),
                                  cert.ref())
        if r:
            raise Error._last(self.context())



from datetime import datetime, timezone

from _sequoia import ffi, lib
from . import error

def invoke(fun, *args):
    """Invokes the given FFI function.

    This function invokes the given FFI function.  It must only be
    used for functions that expect an error pointer as first argument.

    If an error is encountered, an exception is raised.
    """
    err = ffi.new("pgp_error_t[1]")
    result = fun(err, *args)
    if err[0] != ffi.NULL:
        raise Error._from(err[0])
    return result

class SQObject(object):
    # These class attributes determine what features the wrapper class
    # implements.  They must be set to the relevant Sequoia functions.
    #
    # XXX: Once we can assume Python3.6 we can use '__init_subclass__'
    # and reflection on the 'lib' object to set them automatically
    # using the type name.
    _del = None
    _clone = None
    _eq = None
    _str = None
    _debug = None
    _hash = None

    def __init__(self, o, context=None, owner=None, references=None):
        if o == ffi.NULL:
            raise error.Error._last(context)
        self.__o = None
        self.ref_replace(o, owner=owner, references=references)
        self.__ctx = context
        if self.__class__._hash is None and not hasattr(self.__class__, '__hash__'):
            # Unhashable types must have '__hash__' set to None.
            # Until we can use '__init_subclass__', we need to patch
            # the class here.  Yuck.
            self.__class__.__hash__ = None

    def ref(self):
        return self.__o

    def ref_consume(self):
        ref = self.ref()
        self._delete(skip_free=True)
        return ref

    def ref_replace(self, new, owner=None, references=None):
        old = self.ref_consume()
        if self._del and owner == None:
            # There is a destructor and We own the referenced object
            # new.
            self.__o = ffi.gc(new, self._del)
        else:
            self.__o = new
        self.__owner = owner
        self.__references = references
        return old

    def _delete(self, skip_free=False):
        if not self.__o:
            return
        if self._del and skip_free:
            ffi.gc(self.__o, None)
        self.__o = None
        self.__owner = None
        self.__references = None

    def context(self):
        return self.__ctx

    def __str__(self):
        if self._str:
            return _str(self._str(self.ref()))
        else:
            return repr(self)

    def __eq__(self, other):
        if self._eq:
            return (isinstance(other, self.__class__)
                    and bool(self._eq(self.ref(), other.ref())))
        else:
            return NotImplemented

    def copy(self):
        if self._clone:
            return self.__class__(self._clone(self.ref()))
        else:
            raise NotImplementedError()

    def __hash__(self):
        return self._hash(self.ref())

    def debug(self):
        if self._debug:
            return _str(self._debug(self.ref()))
        else:
            raise NotImplementedError()

def sq_str(s):
    t = ffi.string(s).decode()
    lib.free(s)
    return t
_str = sq_str

def sq_static_str(s):
    return ffi.string(s).decode()
_static_str = sq_static_str

def sq_iterator(iterator, next_fn, map=lambda x: x):
    while True:
        entry = next_fn(iterator)
        if entry == ffi.NULL:
            break
        yield map(entry)

def sq_time(t):
    if t == 0:
        return None
    else:
        return datetime.fromtimestamp(t, timezone.utc)



from _sequoia import ffi, lib
from .glue import sq_str

class Error(Exception):
    @classmethod
    def _from(cls, o):
        if o == ffi.NULL:
            return MalformedValue()

        status = lib.pgp_error_status(o)
        return _status_map[status](o)

    @classmethod
    def _last(cls, ctx):
        if not ctx:
            return MalformedValue()
        return Error._from(lib.sq_context_last_error(ctx.ref()))

class MalformedValue(Error, ValueError):
    def __init__(self, message="Malformed value"):
        super(MalformedValue, self).__init__(message)

class SQError(Error):
    def __init__(self, o):
        self.__o = ffi.gc(o, lib.pgp_error_free)
        super(SQError, self).__init__(sq_str(lib.pgp_error_to_string(self.__o)))

class Success(SQError):
    pass

class UnknownError(SQError):
    pass

class NetworkPolicyViolation(SQError):
    pass

class IoError(SQError):
    pass

class InvalidOperataion(SQError):
    pass

class MalformedPacket(SQError):
    pass

class UnsupportedHashAlgorithm(SQError):
    pass

class UnsupportedSymmetricAlgorithm(SQError):
    pass

class InvalidPassword(SQError):
    pass

class InvalidSessionKey(SQError):
    pass

_status_map = {
    lib.PGP_STATUS_SUCCESS: Success,
    lib.PGP_STATUS_UNKNOWN_ERROR: UnknownError,
    lib.PGP_STATUS_NETWORK_POLICY_VIOLATION: NetworkPolicyViolation,
    lib.PGP_STATUS_IO_ERROR: IoError,
    lib.PGP_STATUS_INVALID_OPERATION: InvalidOperataion,
    lib.PGP_STATUS_MALFORMED_PACKET: MalformedPacket,
    lib.PGP_STATUS_UNSUPPORTED_HASH_ALGORITHM: UnsupportedHashAlgorithm,
    lib.PGP_STATUS_UNSUPPORTED_SYMMETRIC_ALGORITHM: UnsupportedSymmetricAlgorithm,
    lib.PGP_STATUS_INVALID_PASSWORD: InvalidPassword,
    lib.PGP_STATUS_INVALID_SESSION_KEY: InvalidSessionKey,
}



import io
from enum import Enum

from _sequoia import ffi, lib
from .error import Error
from .glue import SQObject, invoke

class IPCPolicy(Enum):
    External = lib.SQ_IPC_POLICY_EXTERNAL
    Internal = lib.SQ_IPC_POLICY_INTERNAL
    Robust = lib.SQ_IPC_POLICY_ROBUST

class Context(SQObject):
    _del = lib.sq_context_free
    def __init__(self,
                 home=None,
                 ipc_policy=IPCPolicy.Robust,
                 ephemeral=False):
        cfg = lib.sq_context_configure()
        if home:
            lib.sq_config_home(cfg, home.encode())
        lib.sq_config_ipc_policy(cfg, ipc_policy.value)
        if ephemeral:
            lib.sq_config_ephemeral(cfg)
        err = ffi.new("pgp_error_t[1]")
        o = lib.sq_config_build(cfg, err)
        if o == ffi.NULL:
            raise Error._from(err[0])
        super(Context, self).__init__(o)

class AbstractReader(SQObject, io.RawIOBase):
    _del = lib.pgp_reader_free

    def readable(self):
        return True
    def writable(self):
        return False

    def readinto(self, buf):
        return invoke(
            lib.pgp_reader_read,
            self.ref(),
            ffi.cast("uint8_t *", ffi.from_buffer(buf)),
            len(buf))

    def close(self):
        self._delete()

    # Implement the context manager protocol.
    def __enter__(self):
        return self
    def __exit__(self, *args):
        self.close()
        return False

class Reader(AbstractReader):
    @classmethod
    def open(cls, ctx, filename):
        return Reader(
            invoke(lib.pgp_reader_from_file,
                   filename.encode()),
            context=ctx)

    @classmethod
    def from_fd(cls, ctx, fd):
        return Reader(lib.pgp_reader_from_fd(fd),
                      context=ctx)

    @classmethod
    def from_bytes(cls, ctx, buf):
        return Reader(
            lib.pgp_reader_from_bytes(
                ffi.cast("uint8_t *", ffi.from_buffer(buf)), len(buf)),
            context=ctx)

class AbstractWriter(SQObject, io.RawIOBase):
    _del = lib.pgp_writer_free

    def readable(self):
        return False
    def writable(self):
        return True

    def write(self, buf):
        return invoke(
            lib.pgp_writer_write,
            self.ref(),
            ffi.cast("const uint8_t *", ffi.from_buffer(buf)),
            len(buf))

    def close(self):
        self._delete()

    # Implement the context manager protocol.
    def __enter__(self):
        return self
    def __exit__(self, *args):
        self.close()
        return False

class Writer(AbstractWriter):
    @classmethod
    def open(cls, ctx, filename):
        return Writer(
            invoke(lib.pgp_writer_from_file,
                   filename.encode()),
            context=ctx)

    @classmethod
    def from_fd(cls, ctx, fd):
        return Writer(lib.pgp_writer_from_fd(fd),
                      context=ctx)

    @classmethod
    def from_bytes(cls, ctx, buf):
        return Writer(
            lib.pgp_writer_from_bytes(
                ffi.cast("uint8_t *", ffi.from_buffer(buf)), len(buf)),
            context=ctx)
